/*
 * Copyright 2008, Systemic Realtime Design LLC. All Rights Reserved.
 *
 * This source code may contain confidential information of Systemic
 * Realtime Design LLC and its licensors. Any use, reproduction,
 * modification, disclosure, distribution or transfer of this software,
 * or any software which includes or is based upon any of this code, is
 * prohibited unless expressly authorized by SRD by written agreement. For
 * more information (including whether this source code file has been
 * published) please email licensing@sysrtime.com.
 */

/*
 * net/tipc/tipc_core.c: TIPC module code
 *
 * Copyright (c) 2003-2006, Ericsson AB
 * Copyright (c) 2005-2007, Wind River Systems
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the names of the copyright holders nor the names of its
 *    contributors may be used to endorse or promote products derived from
 *    this software without specific prior written permission.
 *
 * Alternatively, this software may be distributed under the terms of the
 * GNU General Public License ("GPL") version 2 as published by the Free
 * Software Foundation.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include <dlfcn.h>
#include "tipc_core.h"
#include "tipc_dbg.h"
#include "tipc_ref.h"
#include "tipc_net.h"
#include "tipc_user_reg.h"
#include "tipc_name_table.h"
#include "tipc_topsrv.h"
#include "tipc_cfgsrv.h"
#include "tipc_socket.h"
#include "version.h"

/*
 * Declare routines in sub-systems that don't warrant their own .h file
 */

extern int  tipc_handler_start(void);
extern void tipc_handler_stop(void);

extern int  tipc_socket_init(void);
extern void tipc_socket_stop(void);

extern int  tipc_eth_media_start(void);
extern void tipc_eth_media_stop(void);

/* configurable TIPC parameters */

#ifndef CONFIG_TIPC_ADVANCED
#define CONFIG_TIPC_NETID	4711
#define CONFIG_TIPC_REMOTE_MNG	1
#define CONFIG_TIPC_PORTS	8191
#define CONFIG_TIPC_NODES	255
#define CONFIG_TIPC_CLUSTERS	8
#define CONFIG_TIPC_ZONES	4
#define CONFIG_TIPC_REMOTES	8
#define CONFIG_TIPC_PUBL	10000
#define CONFIG_TIPC_SUBSCR	2000
#define CONFIG_TIPC_LOG		0
#else
#ifndef CONFIG_TIPC_REMOTE_MNG
#define CONFIG_TIPC_REMOTE_MNG	0
#endif
#endif

u32 tipc_own_addr;
int tipc_net_id;
int tipc_remote_management;
int tipc_max_nodes;
int tipc_max_clusters;
int tipc_max_zones;
int tipc_max_remotes;
int tipc_max_ports;
int tipc_max_publications;
int tipc_max_subscriptions;

/* global variables used by multiple sub-systems within TIPC */

int tipc_mode = TIPC_NOT_RUNNING;
int tipc_random;
atomic_t tipc_user_count = ATOMIC_INIT(0);

const char tipc_alphabet[] = 
	"ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789_.";


int tipc_get_mode(void)
{
	return tipc_mode;
}

/**
 * buf_acquire - creates a TIPC message buffer
 * @size: message size (including TIPC header)
 *
 * Returns a new buffer with data pointers set to the specified size.
 * 
 * NOTE: Headroom is reserved to allow prepending of a data link header.
 *       There may also be unrequested tailroom present at the buffer's end.
 */

struct sk_buff *buf_acquire(u32 size)
{
	struct sk_buff *skb;
	unsigned int buf_size = (BUF_HEADROOM + size + 3) & ~3u;

	skb = alloc_skb_fclone(buf_size, GFP_ATOMIC);
	if (skb) {
		skb_reserve(skb, BUF_HEADROOM);
		skb_put(skb, size);
		skb->next = NULL;
	}
	dbg_assert(skb != NULL);
	return skb;
}

/**
 * tipc_core_stop_net - shut down TIPC networking sub-systems
 */

void tipc_core_stop_net(void)
{
	tipc_net_stop();
	tipc_eth_media_stop();
}

/**
 * tipc_core_start_net - start TIPC networking sub-systems
 */

int tipc_core_start_net(void)
{
	int res;

	if ((res = tipc_net_start()) ||
	    (res = tipc_eth_media_start())) {
		tipc_core_stop_net();
	}
	return res;
}

/**
 * tipc_core_stop - switch TIPC from SINGLE NODE to NOT RUNNING mode
 */

void tipc_core_stop(void)
{
	if (tipc_mode != TIPC_NODE_MODE)
		return;

	tipc_mode = TIPC_NOT_RUNNING;

	tipc_socket_stop();
	tipc_cfg_stop();
	tipc_subscr_stop();
	tipc_nametbl_stop();
	tipc_routetbl_stop();
	tipc_reg_stop();
	tipc_ref_table_stop();
	tipc_handler_stop();
}

/**
 * tipc_core_start - switch TIPC from NOT RUNNING to SINGLE NODE mode
 */

int tipc_core_start(void)
{
	int res;

	if (tipc_mode != TIPC_NOT_RUNNING)
		return -ENOPROTOOPT;

	tipc_random =
	    (((rand () & 0xFFFF) << 16) | (rand () & 0xFFFF)) ^ tipc_own_addr;
	tipc_mode = TIPC_NODE_MODE;

	if ((res = tipc_handler_start())
	    || (res = tipc_ref_table_init(tipc_max_ports, tipc_random))
	    || (res = tipc_reg_start())
	    || (res = tipc_routetbl_init())
	    || (res = tipc_nametbl_init())
        
		/* MLL@SRD: tipc_subscr_start and tipc_cfg_init can't be run in another thread 
		 *          at this point. io-pkt system is not fully initialized. just call them directly
		 */
        || (res = tipc_subscr_start())
	    || (res = tipc_cfg_init())
	   
	    || (res = tipc_socket_init())
	    ) {
		tipc_core_stop();
	}
	return res;
}

static int tipc_init(void)
{
	int res;
	
	TIPC_LOG->echo = 1;
	tipc_log_resize(CONFIG_TIPC_LOG);
	info("Activated (version " TIPC_MOD_VER 
	     " compiled " __DATE__ " " __TIME__ ")\n");
	info("QNX port version %d.%d.%d\n", VMAJOR, VMINOR, VBUILD); 

	tipc_own_addr = 0;
	tipc_net_id = delimit(CONFIG_TIPC_NETID, 1, 9999);
	tipc_remote_management = CONFIG_TIPC_REMOTE_MNG;
	tipc_max_ports = delimit(CONFIG_TIPC_PORTS, 127, 65536);
	tipc_max_nodes = delimit(CONFIG_TIPC_NODES, 8, 4095);
	tipc_max_clusters = delimit(CONFIG_TIPC_CLUSTERS, 1, 4095);
	tipc_max_zones = delimit(CONFIG_TIPC_ZONES, 1, 255);
	tipc_max_remotes = delimit(CONFIG_TIPC_REMOTES, 0, 255);
	tipc_max_publications = delimit(CONFIG_TIPC_PUBL, 1, 65535);
	tipc_max_subscriptions = delimit(CONFIG_TIPC_SUBSCR, 1, 65535);

	if ((res = tipc_core_start()))
		err("Unable to start in single node mode\n");
	else	
		info("Started in single node mode\n");
	return res;
}


static void tipc_exit(void)
{
	tipc_core_stop_net();
	tipc_core_stop();
	info("Deactivated\n");
	tipc_log_resize(0);
}

/*******************************************************************************
*
* tipc_entry - called by io-pkt via lsm_dll_entry
*
* These functions are called by io-pkt via lsm_dll_entry, they are running
* in io-pkt's context.
*
*/

int optv;
int optb;

static char * const tipc_tokens[] = {
#define TIPC_VERBOSE	0
	"verbose",
#define TIPC_BUFSIZE	1
	"bufsize",
	/* add additional cmd line arguments here */	
	0
};

static int tipc_entry (void *dll_hdl, struct _iopkt_self *iopkt, char *options)
{
	char *options_copy;

    /* make sure this is the right stack */
    if (dlsym(RTLD_DEFAULT, "etype_list_head") == NULL) {
        err("%s(): io-net does not support TIPC.", __FUNCTION__);
        return EINVAL;
    }

	/* defaults */
	optb = TIPC_RX_BUFSIZE;
	optv = 0;

	/* parse stack options */
	if (options)
	{
		options_copy = strdup(options);
		options = options_copy;

		while (options && *options)	
		{
			char *value;
			switch (getsubopt(&options, tipc_tokens, &value))
			{
				case TIPC_VERBOSE:
				if (value)
					optv = strtol(value, 0, 0);
				else
					optv = 1;		
				break;
				
				case TIPC_BUFSIZE:
				if (value)
					optb = strtol(value, 0, 0);
				else
					optb = TIPC_RX_BUFSIZE;
				break;

				default:
					break;	
			}			
		}
		free(options_copy);	
	}
    return tipc_init();
}


struct _iopkt_lsm_entry IOPKT_LSM_ENTRY_SYM(tipc) =
    IOPKT_LSM_ENTRY_SYM_INIT(tipc_entry);


